from dios.lib import *
from dios.options import *
from dios.itypes import *
from dios.errors import *
import pandas as pd
import numpy as np
import operator as op
from functools import partialmethod

from functools import wraps
from collections import OrderedDict
from pandas._libs.lib import to_object_array, is_bool_array
from pandas.core.common import is_bool_indexer
from pandas.core.dtypes.common import (
    is_list_like,
    is_nested_list_like,
    is_scalar,
    is_integer,
    is_dict_like,
    is_number,
)
from pandas.core.dtypes.common import is_iterator as _is_iterator


def is_dios_like(obj):
    return isinstance(obj, DictOfSeries)


def is_pandas_like(obj):
    """We consider ourselfs (dios) as pandas-like"""
    return is_series_like(obj) or is_dataframe_like(obj) or is_dios_like(obj)


def is_series_like(obj):
    return isinstance(obj, pd.Series)


def is_dataframe_like(obj):
    return isinstance(obj, pd.DataFrame)


def is_iterator(obj):
    """ This is only a dummy wrapper, to warn that the docu of this isnt't right.
    Unlike the example says,
    >>is_iterator([1, 2, 3])
    returns False, not True for lists
    """
    return _is_iterator(obj)


class DictOfSeries:
    """
    DictionaryOfSeries is a collection of pd.Series's which aim to be as close as possible similar to
    pd.DataFrame. The advantage over pd.DataFrame is, that every `column` has its own row-index,
    unlike the former, which provide a single row-index for all columns. This solves problems with
    unaligned pd.Series.

    Dataframe vs DictionaryOfSeries
    -------------------------------

    pd.DataFrame :
      - ``df['a']``      indexing with single label, work on columns, return series
      - ``df[iterabel]`` indexing with list, work on columns, return df
      - ``df[slice]``    indexing with slice, work on rows, return df
      - ``df.loc[..]``   loc[cation] based indexing, work on row and/or columns `labels`, return df
      - ``df.iloc[..]``  i[nteger]loc[cation] based indexing, work on row and/or columns, return df

    DictionaryOfSeries:
      - ``dios['a']``       indexing with single label, work on columns, return series
      - ``dios[iterable]``  indexing with list, work on columns, return dios
      - ``dios[slice]``     indexing with slice, work on (all) index'es, return dios
      - ``dios.loc[..]``    loc[cation] based indexing, work on row and/or columns `labels`, return dios
      - ``dios.iloc[..]``   i[nteger]loc[cation] based indexing, work on row and/or columns, return dios

    Todos:
    -----
    todo: to_discuss!! allow any hashable obj as column identifier
         Currently we only allow strings as identifier, to be more df-like we should allow any
         hashable object (unlike df we may should exclude stuff like: ``None`` or ``np.nan`` ??)

    """

    def __init__(self, data=None, columns=None, itype=MixedItype, downcast_policy='lossless'):

        self._data = OrderedDict()

        # We need to keep track of the index-type (itype) of every new Series.
        # If the itypes differ between different series, slicing will almost always fail
        # (eg. a datetime-like slice cannot work on a numeric index and vice versa).
        self._itype = None
        self.itype = get_itype(itype)

        if downcast_policy not in CAST_POLICIES:
            raise ValueError(f"downcast_policy must be one of {CAST_POLICIES}")
        self._policy = downcast_policy

        if data is not None:
            self.__init_insert_data__(data)

        # we use the columns.setter to make all necessary checks
        if columns is not None:
            self.columns = columns

    def __init_insert_data__(self, data):
        if isinstance(data, DictOfSeries):
            g = ((k, data[k]) for k in data)
        else:
            data = list(data) if is_iterator(data) else data
            if is_dict_like(data):
                g = ((k, data[k]) for k in data)
            elif is_nested_list_like(data):
                g = ((str(i), d) for i, d in enumerate(data))
            elif is_list_like(data):
                g = [('0', data)]
            else:
                raise ValueError(f"init with data of type {type(data)} is not possible.")
        for k, val in g:
            self[k] = val
        return

    @property
    def columns(self):
        return list(self._data.keys())

    @columns.setter
    def columns(self, new):
        if not isinstance(new, list):
            raise TypeError("column names must be given as a list")

        if len(set(new)) != len(new):
            raise ValueError("column names must be unique")

        if len(new) != len(self.columns):
            raise ValueError(f"Length mismatch: Columns has {len(self.columns)} elements, "
                             f"new values have {len(new)} elements")

        # to keep order, we iterate over self instead of new
        _d = OrderedDict()
        for i, k in enumerate(self.columns):
            _d[new[i]] = self[k]
        self._data = _d

    @property
    def values(self):
        # will make all series same length, inset nan's
        return to_object_array(self._data.values()).transpose()

    @property
    def data(self):
        return self._data.values()

    @property
    def itype(self):
        return self._itype

    @itype.setter
    def itype(self, itype_like):
        itype = get_itype(itype_like)

        if not itype_le(self._itype, itype):
            self.__cast_all(itype)

        self._itype = itype

        if not itype.unique:
            throw_MixedItype_err_or_warn(f"Using a {itype} as dios.itype is experimental. As soon as series with "
                                         f"different index types are inserted, slicing will almost always fail. "
                                         f"You are hereby warned!")

    def __cast_all(self, itype):
        k = '?'
        try:
            for k in self.columns:
                casted = cast_to_itype(self._data[k], itype, policy=self._policy)
                self._data[k] = casted
        except Exception as e:
            raise type(e)(f"Column {k}: " + str(e)) from e

    def __getitem__(self, key):
        """
        Get items:
          - ``dios['a']``      -> return a pd.Series
          - ``dios[iterable]`` -> return a DictOfSeries of all given columns [1]
          - ``dios[slice]``    -> return a sliced DictOfSeries of all(!) columns in the dios
        Notes:
          - [1] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """
        # special case single label
        if isinstance(key, str):
            if key in self.columns:
                new = self._get_item(key)
            else:
                raise KeyError(key)
        # all other cases
        else:
            keys, ixs = self._get_keys_and_indexer(key)
            new = self.copy_empty()
            for i, _ in enumerate(keys):
                key, ix = keys[i], ixs[i]
                new._data[key] = self._get_item(key)[ix]
        return new

    def _get_item(self, key):
        """Extract a pd.Series from self"""
        return self._data[key]

    def __setitem__(self, key, value):
        """
        Set items:
          - ``dios['a'] = pd.Series()`` -> Set a new pd.Series to an existing column or add it as `new column`.
          - ``dios['a'] = Any`` -> Pass ``any`` to the pd.Series in the corresponding existing(!) column. [1],[2]
          - ``dios[iterable] = Any`` -> Pass ``any`` to the pd.Series's in the corresponding set of columns. [1],[2],[3]
          - ``dios[slice] = Any`` -> Pass ``any`` to all(!) sliced pd.Series's from the current dios. [1],[2]

        Notes:
          - [1] The length of ``Any`` must be be equal to the length of every sliced pd.Series.
          - [2] If ``Any`` is a ``DictOfSeries`` the behavior depends on the option ``dios_to_dios_method``
            in the ``options`` dictionary.
          - [3] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """
        # special case single label
        if isinstance(key, str):
            if key not in self.columns:
                self._insert(key, value)
                return
            else:
                k, i = [key], [slice(None)]
        # all other cases
        else:
            k, i = self._get_keys_and_indexer(key)

        gen = self._yield_tuple_to_set(k, i, value)
        for tup in gen:
            self._set_item(*tup)

    def _set_item(self, key, ix, val):
        "Set a value (scalar or list or series)"
        ser = self._data[key]
        if is_series_like(val):
            left = ser[ix]
            index = left.index.intersection(val.index)
            if not index.empty:
                left.loc[index] = val.loc[index].copy()
        else:
            ser[ix] = val

    def _yield_tuple_to_set(self, keys, ixs, val):
        """Return a generator that yield (key, indexer, value) for all keys"""
        val = list(val) if is_iterator(val) else val
        diosl, dfl, nlistl = is_dios_like(val), is_dataframe_like(val), is_nested_list_like(val)

        if (diosl or dfl or nlistl) and len(val) != len(keys):
            raise ValueError(f"could not broadcast input array with length {len(val)}"
                             f" into dios of length {len(keys)}")

        # now we have everything we need: key, indexer, value
        # so we just pack it nice and cosy and let setitem
        # do the dirty work.
        for i, _ in enumerate(keys):
            key, ix = keys[i], ixs[i]
            if dfl or diosl:
                yield key, ix, val[val.columns[i]]
            elif nlistl:
                yield key, ix, val[i]
            else:
                yield key, ix, val

    def _insert(self, key, val):
        """Insert a fresh new value into self"""
        if isinstance(val, DictOfSeries):
            val = val.squeeze()
        elif is_list_like(val) and not is_nested_list_like(val):
            val = pd.Series(val)

        if not isinstance(val, pd.Series):
            raise ValueError(f"Only pd.Series can be inserted directly")

        val = cast_to_itype(val, self._itype, policy=self._policy)
        self._data[key] = val.copy(deep=True)

    def _get_keys_and_indexer(self, key):
        """ Determine keys and indexer by type of key. This does not deal
            with single label-access, only higher dimension objects are handled..

        Notes:
            Which keys we get, may depend on the policy in dios_options
        """
        err_bool = "only boolen values are allowed"
        keys = None
        indexers = None
        blowup = False

        # prevent consuming of a generator
        key = list(key) if is_iterator(key) else key

        if isinstance(key, slice):
            keys = self.columns
            indexers, blowup = [key], True

        # list, np.arrays, ... of list, np.arrays..
        elif is_nested_list_like(key):
            keys = self.columns
            indexers = []
            # we only allow nested lists with bool entries
            for i in range(len(key)):
                arr = np.array(i)
                if not is_bool_array(arr):
                    raise ValueError("nested list: " + err_bool)
                indexers.append(arr)

        # ser, df, dios
        elif is_pandas_like(key):
            if is_series_like(key):
                keys = key.to_list()

            elif is_dataframe_like(key):
                keys = key.columns.to_list()
                indexers = key.values
                if not is_bool_array(indexers):
                    raise ValueError("df: " + err_bool)

            elif is_dios_like(key):
                keys = key.columns
                indexers = list(key.values)
                if not is_bool_array(indexers):
                    raise ValueError("dios: " + err_bool)

        # list, np.array, np.ndarray, ...
        # Note: series considered list-like,
        # so we handle lists last
        elif is_list_like(key):
            arr = np.array(key)
            if is_bool_array(arr):
                keys = self.columns
                if len(arr) != len(keys):
                    keys = np.array(keys)[arr]
            else:
                keys = key

        else:
            raise KeyError(f"{key}")

        # check keys
        method = dios_options[Options.dios_to_dios_method]
        keys = check_keys_by_policy(keys, self.columns, method)

        # check indexer
        if indexers is None:
            indexers, blowup = [slice(None)], True
        if blowup:
            indexers = indexers * len(keys)
        if len(indexers) != len(keys):
            raise ValueError

        # now we have a valid indexer (a slice or a bool array) for every series
        return keys, indexers

    @property
    def loc(self):
        return _LocIndexer(self)

    @property
    def iloc(self):
        return _iLocIndexer(self)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        pd_max_rows = pd.get_option('display.max_rows')
        pd.set_option('display.max_rows', dios_options[Options.disp_max_rows])

        def item2str(k):
            kstr = str(self[k]).replace('\n', '\n ')
            return f'{k}:\n {kstr}\n'

        maxrows = dios_options[Options.disp_max_vars]
        s = ''
        head = maxrows // 2
        tail = len(self.columns) - head
        for i, k in enumerate(self.columns):
            if i < head:
                s += item2str(k)
            elif i == head:
                s += ' ... \t\t\t\t ...\n' * 2
            elif i > tail:
                s += item2str(k)
        pd.set_option('display.max_rows', pd_max_rows)
        return s

    @property
    def empty(self):
        return len(self) == 0 or all(self._data[c].empty for c in self._data)

    def all(self):
        return pd.Series([self._data[c].any() for c in self._data])

    def any(self):
        return pd.Series([self._data[c].any() for c in self._data])

    def aaall(self):
        """absolute all all"""
        return self.all().all()

    def aaany(self):
        """absolute any any"""
        return self.any().any()

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        yield from self._data

    def __reversed__(self):
        yield from reversed(self._data)

    def __contains__(self, item):
        return item in self._data

    def __delitem__(self, key):
        del self._data[key]

    def __copy__(self):
        return self.copy(deep=True)

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    def copy(self, deep=True):
        new = self.copy_empty()
        # We use `_data` here, because all checks are already done.
        # So this should be much faster, especially, because we use the underlying dict for
        # getting and setting the values, instead of ``__setitem__`` and ``__getitem__``.
        for k in self._data:
            new._data[k] = self._data[k].copy(deep=deep)
        return new

    def copy_empty(self):
        new = DictOfSeries()
        new._itype = self.itype
        return new

    def _op1(self, op):
        new = self.copy_empty()
        try:
            for k in self:
                new[k] = op(self._data[k])
        except Exception as e:
            raise type(e)(f"'{OP_MAP[op]} dios' failed: " + str(e)) from e
        return new

    def _op2(self, op, other, inplace=False):
        def raiseif(cond, s='lenght'):
            if cond:
                raise ValueError(f"{s} does not match, {s} left: {len(self)}, {s} right: {len(other)} keys")

        def gen():
            if isinstance(other, (self.__class__, pd.DataFrame)):
                raiseif(set(other) != set(self), '#keys')
                for k in self.columns:
                    left, right = self._data[k], other[k]
                    yield k, op(*(left.align(right, join='inner')))
            elif isinstance(other, pd.Series):
                for k in self.columns:
                    left, right = self._data[k], other
                    yield k, op(*(left.align(right, join='inner')))
            elif is_dict_like(other):
                raiseif(set(other) != set(self), '#keys')
                for k in self.columns:
                    yield k, op(self._data[k], other[k])
            elif is_nested_list_like(other):
                raiseif(len(other) != len(self), 'length')
                for i, k in enumerate(self.columns):
                    yield k, op(self._data[k], other[i])
            elif is_scalar(other) or is_list_like(other):
                for k in self.columns:
                    yield k, op(self._data[k], other)
            else:
                raise NotImplementedError

        new = self if inplace else self.copy_empty()
        try:
            for k, val in gen():
                new[k] = val
        except Exception as e:
            raise type(e)(f"'dios {OP_MAP[op]} other' failed: " + str(e)) from e
        return new

    __neg__ = partialmethod(_op1, op.neg)
    __abs__ = partialmethod(_op1, op.abs)
    __invert__ = partialmethod(_op1, op.inv)
    __eq__ = partialmethod(_op2, op.eq)
    __ne__ = partialmethod(_op2, op.ne)
    __le__ = partialmethod(_op2, op.le)
    __ge__ = partialmethod(_op2, op.ge)
    __lt__ = partialmethod(_op2, op.lt)
    __gt__ = partialmethod(_op2, op.gt)
    __add__ = partialmethod(_op2, op.add)
    __sub__ = partialmethod(_op2, op.sub)
    __mul__ = partialmethod(_op2, op.mul)
    __mod__ = partialmethod(_op2, op.mod)
    __truediv__ = partialmethod(_op2, op.truediv)
    __floordiv__ = partialmethod(_op2, op.floordiv)
    __pow__ = partialmethod(_op2, op.pow)
    __and__ = partialmethod(_op2, op.and_)
    __or__ = partialmethod(_op2, op.or_)
    __xor__ = partialmethod(_op2, op.xor)

    def squeeze(self):
        return self[self.columns[0]] if len(self) == 1 else self

    def memory_usage(self, index=True, deep=False):
        mem = 0
        for k in self.columns:
            mem += self[k].memory_usage(index=index, deep=deep)
        return mem

    def foreach(self, f, *args, **kwargs):
        self.pipe(f, *args, **kwargs)

    def pipe(self, f, *args, **kwargs):
        """
        Call a pd.Series function on all series in the DictOfSeries

        :param f: pd.Series function to apply on all Series
        :param args: args for the function
        :param kwargs: kwargs for the function
        :return:
        """
        news = pd.Series()
        newd = self.copy_empty()
        need_dios = False
        # return a copy nevertheless, but also run inplace if inplace=True or
        # if the function not has this option, but work inplace.

        inplace = kwargs.get('inplace', False) or kwargs.get('copy', False)

        def getobj(k, inplace):
            if inplace:
                # self[k] may return a copy, so we need to get
                # the item from the dict, by our mother class.
                return self._data[k]
            else:
                return self._data[k].copy()

        for k in self.columns:
            v = f(getobj(k, inplace), *args, **kwargs)
            if inplace:
                continue
            if isinstance(v, pd.Series):
                newd[k] = v
                need_dios = True
            else:
                newd[k] = pd.Series(data=v)
                news[k] = v

        if inplace:
            return None
        if need_dios:
            return newd.squeeze()
        if len(news) == 0:
            return None
        return news.squeeze()


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        self._data = _dios._data
        self._yield_tuple_to_set = _dios._yield_tuple_to_set



class _LocIndexer(_Indexer):

    def __init__(self, _dios):
        super().__init__(_dios)
        self._set_item = _dios._set_item
        self._get_keys_and_indexer = _dios._get_keys_and_indexer

    def __getitem__(self, key):
        rkey, cols = self._unpack_key(key)
        new = self._dios.copy_empty()
        for c in cols:
            new[c] = self._data[c].loc[rkey]
        return new

    def __setitem__(self, key, value):
        ixs, keys = self._unpack_key(key)
        gen = self._yield_tuple_to_set(keys, ixs, value)
        for tup in gen:
            self._set_item(*tup)

    def _unpack_key(self, key):
        # if we have a tuple, we have a rows- and a column-indexer
        # if not, we only have a row-indexer and work on all columns
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            ckey = list(ckey) if is_iterator(ckey) else ckey

            # determine columns
            if is_dataframe_like(ckey) or is_nested_list_like(ckey) or is_dios_like(ckey):
                raise ValueError("Cannot index with multidimensional key")
            if isinstance(ckey, str):
                cols = [ckey]
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            else:
                try:
                    # list and bool list like
                    cols, _ = self._get_keys_and_indexer(key)
                except Exception:
                    raise
        else:
            cols = self._data.keys()
            rkey = key
        # blowup
        rkey = [rkey] * len(cols)
        return rkey, cols

    def _col_slice_to_col_list(self, cslice):
        """ see here:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#indexing-slicing-with-labels
        """
        keys = list(self._data.keys())
        try:
            start = keys.index(cslice.start) if cslice.start is not None else None
            stop = keys.index(cslice.stop) if cslice.stop is not None else None
        except ValueError:
            raise KeyError("The slice start label, or the slice stop label, is not present in columns.")
        if not is_integer(cslice.step) or cslice.step <= 0:
            return []
        return keys[slice(start, stop + 1, cslice.step)]


class _iLocIndexer(_Indexer):

    def __getitem__(self, key):
        rkey, cols = self._unpack_key(key)
        new = self._dios.copy_empty()
        for c in cols:
            new[c] = self._data[c].iloc[rkey]
        return new

    def __setitem__(self, key, value):
        ixs, keys = self._unpack_key(key)
        gen = self._yield_tuple_to_set(keys, ixs, value)
        for tup in gen:
            self._set_item_positional(*tup)
        raise NotImplemented

    def _set_item_positional(self, key, ix, val):
        ser = self._data[key]
        if is_series_like(val):
            index = ser.iloc[ix].index
            index = index.intersection(val.index)
            if not index.empty:
                ser.loc[index] = val.loc[index].copy()
        else:
            ser.iloc[ix] = val



    def _unpack_key(self, key):
        # if we have a tuple, we have a rows- and a column-indexer
        # if not, we only have a row-indexer and work on all columns
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            ckey = list(ckey) if is_iterator(ckey) else ckey

            # determine columns
            if is_integer(ckey):
                self._check_keys([ckey])
                cols = self._integers_to_col_list([ckey])
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            elif is_list_like(ckey) and not is_nested_list_like(ckey):
                arr = np.array(ckey)
                if is_bool_array(arr):
                    raise NotImplementedError
                self._check_keys(ckey)
                cols = self._integers_to_col_list(ckey)
            elif is_series_like(ckey):
                raise NotImplementedError
            elif is_bool_indexer(ckey):
                raise NotImplementedError
            else:
                raise KeyError(f"{ckey} of type {type(ckey)}")
        else:
            cols = self._data.keys()
            rkey = key
        return rkey, cols

    def _check_keys(self, keys):
        bound = len(self._data)
        for k in keys:
            if not is_integer(k):
                raise ValueError(f"{type(k)} is not integer")
            if k not in range(-bound, bound):
                raise KeyError("positional indexer(s) are out-of-bounds in columns")

    def _integers_to_col_list(self, ints):
        klist = list(self._data.keys())
        ks = set()
        for i in ints:
            ks.add(klist[i])
        return ks

    def _col_slice_to_col_list(self, sl):
        for s in [sl.start, sl.stop, sl.step]:
            if not is_integer(s):
                raise TypeError(f"positional indexing with slice must be integers, passed type was {type(s)}")
        return list(self._data.keys())[sl]
